@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     pint.util
     ~~~~~~~~~
@@ -9,48 +8,72 @@
     :license: BSD, see LICENSE for more details.
 """
 
-from __future__ import division, unicode_literals, print_function, absolute_import
-
-from decimal import Decimal
-import locale
-import sys
-import re
-import operator
-from numbers import Number
-from fractions import Fraction
-from collections import Mapping
-
 import logging
-from token import STRING, NAME, OP, NUMBER
-from tokenize import untokenize
+import math
+import operator
+import re
+from collections.abc import Mapping
+from decimal import Decimal
+from fractions import Fraction
+from functools import lru_cache
+from logging import NullHandler
+from numbers import Number
+from token import NAME, NUMBER
 
-from .compat import string_types, tokenizer, lru_cache, NullHandler, maketrans, NUMERIC_TYPES
-from .formatting import format_unit,siunitx_format_unit
-from .pint_eval import build_eval_tree
+from .compat import NUMERIC_TYPES, tokenizer
 from .errors import DefinitionSyntaxError
+from .formatting import format_unit
+from .pint_eval import build_eval_tree
 
 logger = logging.getLogger(__name__)
 logger.addHandler(NullHandler())
 
 
-def matrix_to_string(matrix, row_headers=None, col_headers=None, fmtfun=lambda x: str(int(x))):
+def matrix_to_string(
+    matrix, row_headers=None, col_headers=None, fmtfun=lambda x: str(int(x))
+):
     """Takes a 2D matrix (as nested list) and returns a string.
+
+    Parameters
+    ----------
+    matrix :
+
+    row_headers :
+         (Default value = None)
+    col_headers :
+         (Default value = None)
+    fmtfun :
+         (Default value = lambda x: str(int(x)))
+
+    Returns
+    -------
+
     """
     ret = []
     if col_headers:
-        ret.append(('\t' if row_headers else '') + '\t'.join(col_headers))
+        ret.append(("\t" if row_headers else "") + "\t".join(col_headers))
     if row_headers:
-        ret += [rh + '\t' + '\t'.join(fmtfun(f) for f in row)
-                for rh, row in zip(row_headers, matrix)]
+        ret += [
+            rh + "\t" + "\t".join(fmtfun(f) for f in row)
+            for rh, row in zip(row_headers, matrix)
+        ]
     else:
-        ret += ['\t'.join(fmtfun(f) for f in row)
-                for row in matrix]
+        ret += ["\t".join(fmtfun(f) for f in row) for row in matrix]
 
-    return '\n'.join(ret)
+    return "\n".join(ret)
 
 
 def transpose(matrix):
     """Takes a 2D matrix (as nested list) and returns the transposed version.
+
+    Parameters
+    ----------
+    matrix :
+
+
+    Returns
+    -------
+
     """
     return [list(val) for val in zip(*matrix)]
 
@@ -58,10 +81,20 @@ def transpose(matrix):
 def column_echelon_form(matrix, ntype=Fraction, transpose_result=False):
     """Calculates the column echelon form using Gaussian elimination.
 
-    :param matrix: a 2D matrix as nested list.
-    :param ntype: the numerical type to use in the calculation.
-    :param transpose_result: indicates if the returned matrix should be transposed.
-    :return: column echelon form, transformed identity matrix, swapped rows
+    Parameters
+    ----------
+    matrix :
+        a 2D matrix as nested list.
+    ntype :
+        the numerical type to use in the calculation. (Default value = Fraction)
+    transpose_result :
+        indicates if the returned matrix should be transposed. (Default value = False)
+
+    Returns
+    -------
+    type
+        column echelon form, transformed identity matrix, swapped rows
+
     """
     lead = 0
 
@@ -83,8 +116,10 @@ def column_echelon_form(matrix, ntype=Fraction, transpose_result=False):
         new_M.append(r)
     M = new_M
 
-#    M = [[ntype(x) for x in row] for row in M]
-    I = [[ntype(1) if n == nc else ntype(0) for nc in range(rows)] for n in range(rows)]
+    # M = [[ntype(x) for x in row] for row in M]
+    I = [  # noqa: E741
+        [ntype(1) if n == nc else ntype(0) for nc in range(rows)] for n in range(rows)
+    ]
     swapped = []
 
     for r in range(rows):
@@ -112,8 +147,8 @@ def column_echelon_form(matrix, ntype=Fraction, transpose_result=False):
             if i == r:
                 continue
             lv = M[i][lead]
-            M[i] = [iv - lv*rv for rv, iv in zip(M[r], M[i])]
-            I[i] = [iv - lv*rv for rv, iv in zip(I[r], I[i])]
+            M[i] = [iv - lv * rv for rv, iv in zip(M[r], M[i])]
+            I[i] = [iv - lv * rv for rv, iv in zip(I[r], I[i])]
 
         lead += 1
 
@@ -123,9 +158,18 @@ def column_echelon_form(matrix, ntype=Fraction, transpose_result=False):
 def pi_theorem(quantities, registry=None):
     """Builds dimensionless quantities using the Buckingham π theorem
 
-    :param quantities: mapping between variable name and units
-    :type quantities: dict
-    :return: a list of dimensionless quantities expressed as dicts
+    Parameters
+    ----------
+    quantities : dict
+        mapping between variable name and units
+    registry :
+         (Default value = None)
+
+    Returns
+    -------
+    type
+        a list of dimensionless quantities expressed as dicts
+
     """
 
     # Preprocess input and build the dimensionality Matrix
@@ -138,18 +182,20 @@ def pi_theorem(quantities, registry=None):
         getdim = registry.get_dimensionality
 
     for name, value in quantities.items():
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             value = ParserHelper.from_string(value)
         if isinstance(value, dict):
             dims = getdim(UnitsContainer(value))
-        elif not hasattr(value, 'dimensionality'):
+        elif not hasattr(value, "dimensionality"):
             dims = getdim(value)
         else:
             dims = value.dimensionality
 
-        if not registry and any(not key.startswith('[') for key in dims):
-            logger.warning('A non dimension was found and a registry was not provided. '
-                           'Assuming that it is a dimension name: {0}.'.format(dims))
+        if not registry and any(not key.startswith("[") for key in dims):
+            logger.warning(
+                "A non dimension was found and a registry was not provided. "
+                "Assuming that it is a dimension name: {}.".format(dims)
+            )
 
         quant.append((name, dims))
         dimensions = dimensions.union(dims.keys())
@@ -157,8 +203,10 @@ def pi_theorem(quantities, registry=None):
     dimensions = list(dimensions)
 
     # Calculate dimensionless  quantities
-    M = [[dimensionality[dimension] for name, dimensionality in quant]
-         for dimension in dimensions]
+    M = [
+        [dimensionality[dimension] for name, dimensionality in quant]
+        for dimension in dimensions
+    ]
 
     M, identity, pivot = column_echelon_form(M, transpose_result=False)
 
@@ -171,40 +219,54 @@ def pi_theorem(quantities, registry=None):
             continue
         max_den = max(f.denominator for f in rowi)
         neg = -1 if sum(f < 0 for f in rowi) > sum(f > 0 for f in rowi) else 1
-        results.append(dict((q[0], neg * f.numerator * max_den / f.denominator)
-                            for q, f in zip(quant, rowi) if f.numerator != 0))
+        results.append(
+            dict(
+                (q[0], neg * f.numerator * max_den / f.denominator)
+                for q, f in zip(quant, rowi)
+                if f.numerator != 0
+            )
+        )
     return results
 
 
 def solve_dependencies(dependencies):
     """Solve a dependency graph.
 
-    :param dependencies: dependency dictionary. For each key, the value is
-                         an iterable indicating its dependencies.
-    :return: list of sets, each containing keys of independents tasks dependent
-                           only of the previous tasks in the list.
+    Parameters
+    ----------
+    dependencies :
+        dependency dictionary. For each key, the value is an iterable indicating its
+        dependencies.
+
+    Returns
+    -------
+    type
+        iterator of sets, each containing keys of independents tasks dependent only of
+        the previous tasks in the list.
+
     """
-    d = dict((key, set(dependencies[key])) for key in dependencies)
-    r = []
-    while d:
+    while dependencies:
         # values not in keys (items without dep)
-        t = set(i for v in d.values() for i in v) - set(d.keys())
+        t = {i for v in dependencies.values() for i in v} - dependencies.keys()
         # and keys without value (items without dep)
-        t.update(k for k, v in d.items() if not v)
+        t.update(k for k, v in dependencies.items() if not v)
         # can be done right away
         if not t:
-            raise ValueError('Cyclic dependencies exist among these items: {}'.format(', '.join(repr(x) for x in d.items())))
-        r.append(t)
+            raise ValueError(
+                "Cyclic dependencies exist among these items: {}".format(
+                    ", ".join(repr(x) for x in dependencies.items())
+                )
+            )
         # and cleaned up
-        d = dict(((k, v - t) for k, v in d.items() if v))
-    return r
+        dependencies = {k: v - t for k, v in dependencies.items() if v}
+        yield t
 
 
 def find_shortest_path(graph, start, end, path=None):
     path = (path or []) + [start]
     if start == end:
         return path
-    if not start in graph:
+    if start not in graph:
         return None
     shortest = None
     for node in graph[start]:
@@ -217,10 +279,10 @@ def find_shortest_path(graph, start, end, path=None):
 
 
 def find_connected_nodes(graph, start, visited=None):
-    if not start in graph:
+    if start not in graph:
         return None
 
-    visited = (visited or set())
+    visited = visited or set()
     visited.add(start)
 
     for node in graph[start]:
@@ -231,11 +293,13 @@ def find_connected_nodes(graph, start, visited=None):
 
 
 class udict(dict):
-    """ Custom dict implementing __missing__.
+    """Custom dict implementing __missing__."""
 
-    """
     def __missing__(self, key):
-        return 0.
+        return 0.0
+
+    def copy(self):
+        return udict(self)
 
 
 class UnitsContainer(Mapping):
@@ -243,22 +307,30 @@ class UnitsContainer(Mapping):
     exponent and implements the corresponding operations.
 
     UnitsContainer is a read-only mapping. All operations (even in place ones)
-    return new instances.
+
+    Parameters
+    ----------
+
+    Returns
+    -------
+    type
+
 
     """
-    __slots__ = ('_d', '_hash')
+
+    __slots__ = ("_d", "_hash")
 
     def __init__(self, *args, **kwargs):
         d = udict(*args, **kwargs)
         self._d = d
         for key, value in d.items():
-            if not isinstance(key, string_types):
-                raise TypeError('key must be a str, not {0}'.format(type(key)))
+            if not isinstance(key, str):
+                raise TypeError("key must be a str, not {}".format(type(key)))
             if not isinstance(value, Number):
-                raise TypeError('value must be a number, not {0}'.format(type(value)))
+                raise TypeError("value must be a number, not {}".format(type(value)))
             if not isinstance(value, float):
                 d[key] = float(value)
-        self._hash = hash(frozenset(self._d.items()))
+        self._hash = None
 
     def copy(self):
         return self.__copy__()
@@ -269,24 +341,46 @@ class UnitsContainer(Mapping):
         if newval:
             new._d[key] = newval
         else:
-            del new._d[key]
-
+            new._d.pop(key)
+        new._hash = None
         return new
 
     def remove(self, keys):
-        """ Create a new UnitsContainer purged from given keys.
+        """Create a new UnitsContainer purged from given keys.
+
+        Parameters
+        ----------
+        keys :
+
+
+        Returns
+        -------
 
         """
-        d = udict(self._d)
-        return UnitsContainer(((key, d[key]) for key in d if key not in keys))
+        new = self.copy()
+        for k in keys:
+            new._d.pop(k)
+        new._hash = None
+        return new
 
     def rename(self, oldkey, newkey):
-        """ Create a new UnitsContainer in which an entry has been renamed.
+        """Create a new UnitsContainer in which an entry has been renamed.
+
+        Parameters
+        ----------
+        oldkey :
+
+        newkey :
+
+
+        Returns
+        -------
 
         """
-        d = udict(self._d)
-        d[newkey] = d.pop(oldkey)
-        return UnitsContainer(d)
+        new = self.copy()
+        new._d[newkey] = new._d.pop(oldkey)
+        new._hash = None
+        return new
 
     def __iter__(self):
         return iter(self._d)
@@ -301,31 +395,35 @@ class UnitsContainer(Mapping):
         return key in self._d
 
     def __hash__(self):
+        if self._hash is None:
+            self._hash = hash(frozenset(self._d.items()))
         return self._hash
 
-    def __getstate__(self):
-        return {'_d': self._d, '_hash': self._hash}
-
-    def __setstate__(self, state):
-        self._d = state['_d']
-        self._hash = state['_hash']
-
     def __eq__(self, other):
         if isinstance(other, UnitsContainer):
+            # UnitsContainer.__hash__(self) is not the same as hash(self); see
+            # ParserHelper.__hash__ and __eq__.
+            # Different hashes guarantee that the actual contents are different, but
+            # identical hashes give no guarantee of equality.
+            # e.g. in CPython, hash(-1) == hash(-2)
+            if UnitsContainer.__hash__(self) != UnitsContainer.__hash__(other):
+                return False
             other = other._d
-        elif isinstance(other, string_types):
+
+        elif isinstance(other, str):
             other = ParserHelper.from_string(other)
             other = other._d
 
         return dict.__eq__(self._d, other)
 
     def __str__(self):
-        return self.__format__('')
+        return self.__format__("")
 
     def __repr__(self):
-        tmp = '{%s}' % ', '.join(["'{0}': {1}".format(key, value)
-                                  for key, value in sorted(self._d.items())])
-        return '<UnitsContainer({0})>'.format(tmp)
+        tmp = "{%s}" % ", ".join(
+            ["'{}': {}".format(key, value) for key, value in sorted(self._d.items())]
+        )
+        return "<UnitsContainer({})>".format(tmp)
 
     def __format__(self, spec):
         return format_unit(self, spec)
@@ -334,73 +432,83 @@ class UnitsContainer(Mapping):
         return format_unit(self, spec, **kwspec)
 
     def __copy__(self):
-        return UnitsContainer(self._d)
+        # Skip expensive health checks performed by __init__
+        out = object.__new__(self.__class__)
+        out._d = self._d.copy()
+        out._hash = self._hash
+        return out
 
     def __mul__(self, other):
-        d = udict(self._d)
         if not isinstance(other, self.__class__):
-            err = 'Cannot multiply UnitsContainer by {0}'
+            err = "Cannot multiply UnitsContainer by {}"
             raise TypeError(err.format(type(other)))
-        for key, value in other.items():
-            d[key] += value
-        keys = [key for key, value in d.items() if value == 0]
-        for key in keys:
-            del d[key]
 
-        return UnitsContainer(d)
+        new = self.copy()
+        for key, value in other.items():
+            new._d[key] += value
+            if new._d[key] == 0:
+                del new._d[key]
+
+        new._hash = None
+        return new
 
     __rmul__ = __mul__
 
     def __pow__(self, other):
         if not isinstance(other, NUMERIC_TYPES):
-            err = 'Cannot power UnitsContainer by {0}'
+            err = "Cannot power UnitsContainer by {}"
             raise TypeError(err.format(type(other)))
-        d = udict(self._d)
-        for key, value in d.items():
-            d[key] *= other
-        return UnitsContainer(d)
+
+        new = self.copy()
+        for key, value in new._d.items():
+            new._d[key] *= other
+        new._hash = None
+        return new
 
     def __truediv__(self, other):
         if not isinstance(other, self.__class__):
-            err = 'Cannot divide UnitsContainer by {0}'
+            err = "Cannot divide UnitsContainer by {}"
             raise TypeError(err.format(type(other)))
 
-        d = udict(self._d)
-
+        new = self.copy()
         for key, value in other.items():
-            d[key] -= value
+            new._d[key] -= value
+            if new._d[key] == 0:
+                del new._d[key]
 
-        keys = [key for key, value in d.items() if value == 0]
-        for key in keys:
-            del d[key]
-
-        return UnitsContainer(d)
+        new._hash = None
+        return new
 
     def __rtruediv__(self, other):
         if not isinstance(other, self.__class__) and other != 1:
-            err = 'Cannot divide {0} by UnitsContainer'
+            err = "Cannot divide {} by UnitsContainer"
             raise TypeError(err.format(type(other)))
 
-        return self**-1
+        return self ** -1
 
 
 class ParserHelper(UnitsContainer):
-    """ The ParserHelper stores in place the product of variables and
+    """The ParserHelper stores in place the product of variables and
     their respective exponent and implements the corresponding operations.
 
     ParserHelper is a read-only mapping. All operations (even in place ones)
-    return new instances.
 
-    WARNING : The hash value used does not take into account the scale
-    attribute so be careful if you use it as a dict key and then two unequal
-    object can have the same hash.
+    Parameters
+    ----------
+
+    Returns
+    -------
+    type
+        WARNING : The hash value used does not take into account the scale
+        attribute so be careful if you use it as a dict key and then two unequal
+        object can have the same hash.
 
     """
 
-    __slots__ = ('scale', )
+    __slots__ = ("scale",)
 
     def __init__(self, scale=1, *args, **kwargs):
-        super(ParserHelper, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.scale = scale
 
     @classmethod
@@ -409,13 +517,17 @@ class ParserHelper(UnitsContainer):
 
         Equivalent to: ParserHelper({'word': 1})
 
+        Parameters
+        ----------
+        input_word :
+
+
+        Returns
+        -------
+
         """
         return cls(1, [(input_word, 1)])
 
-    @classmethod
-    def from_string(cls, input_string):
-        return cls._from_string(input_string)
-    
     @classmethod
     def eval_token(cls, token, use_decimal=False):
         token_type = token.type
@@ -430,20 +542,30 @@ class ParserHelper(UnitsContainer):
         elif token_type == NAME:
             return ParserHelper.from_word(token_text)
         else:
-            raise Exception('unknown token type')
-    
+            raise Exception("unknown token type")
+
     @classmethod
     @lru_cache()
-    def _from_string(cls, input_string):
+    def from_string(cls, input_string):
         """Parse linear expression mathematical units and return a quantity object.
 
+        Parameters
+        ----------
+        input_string :
+
+
+        Returns
+        -------
+
         """
         if not input_string:
             return cls()
 
         input_string = string_preprocessor(input_string)
-        if '[' in input_string:
-            input_string = input_string.replace('[', '__obra__').replace(']', '__cbra__')
+        if "[" in input_string:
+            input_string = input_string.replace("[", "__obra__").replace(
+                "]", "__cbra__"
+            )
             reps = True
         else:
             reps = False
@@ -454,35 +576,45 @@ class ParserHelper(UnitsContainer):
         if isinstance(ret, Number):
             return ParserHelper(ret)
 
-        if not reps:
-            return ret
+        if reps:
+            ret = ParserHelper(
+                ret.scale,
+                {
+                    key.replace("__obra__", "[").replace("__cbra__", "]"): value
+                    for key, value in ret.items()
+                },
+            )
 
-        return ParserHelper(ret.scale,
-                            dict((key.replace('__obra__', '[').replace('__cbra__', ']'), value)
-                                 for key, value in ret.items()))
+        for k in list(ret):
+            if k.lower() == "nan":
+                del ret._d[k]
+                ret.scale = math.nan
+
+        return ret
 
     def __copy__(self):
-        return ParserHelper(scale=self.scale, **self)
+        new = super().__copy__()
+        new.scale = self.scale
+        return new
 
     def copy(self):
         return self.__copy__()
 
     def __hash__(self):
         if self.scale != 1.0:
-            mess = 'Only scale 1.0 ParserHelper instance should be considered hashable'
+            mess = "Only scale 1.0 ParserHelper instance should be considered hashable"
             raise ValueError(mess)
-        return self._hash
+        return super().__hash__()
 
     def __eq__(self, other):
-        if isinstance(other, self.__class__):
-            return self.scale == other.scale and\
-                super(ParserHelper, self).__eq__(other)
-        elif isinstance(other, string_types):
+        if isinstance(other, ParserHelper):
+            return self.scale == other.scale and super().__eq__(other)
+        elif isinstance(other, str):
             return self == ParserHelper.from_string(other)
         elif isinstance(other, Number):
             return self.scale == other and not len(self._d)
         else:
-            return self.scale == 1. and super(ParserHelper, self).__eq__(other)
+            return self.scale == 1.0 and super().__eq__(other)
 
     def operate(self, items, op=operator.iadd, cleanup=True):
         d = udict(self._d)
@@ -497,17 +629,19 @@ class ParserHelper(UnitsContainer):
         return self.__class__(self.scale, d)
 
     def __str__(self):
-        tmp = '{%s}' % ', '.join(["'{0}': {1}".format(key, value)
-                                  for key, value in sorted(self._d.items())])
-        return '{0} {1}'.format(self.scale, tmp)
+        tmp = "{%s}" % ", ".join(
+            ["'{}': {}".format(key, value) for key, value in sorted(self._d.items())]
+        )
+        return "{} {}".format(self.scale, tmp)
 
     def __repr__(self):
-        tmp = '{%s}' % ', '.join(["'{0}': {1}".format(key, value)
-                                  for key, value in sorted(self._d.items())])
-        return '<ParserHelper({0}, {1})>'.format(self.scale, tmp)
+        tmp = "{%s}" % ", ".join(
+            ["'{}': {}".format(key, value) for key, value in sorted(self._d.items())]
+        )
+        return "<ParserHelper({}, {})>".format(self.scale, tmp)
 
     def __mul__(self, other):
-        if isinstance(other, string_types):
+        if isinstance(other, str):
             new = self.add(other, 1)
         elif isinstance(other, Number):
             new = self.copy()
@@ -525,10 +659,10 @@ class ParserHelper(UnitsContainer):
         d = self._d.copy()
         for key in self._d:
             d[key] *= other
-        return self.__class__(self.scale**other, d)
+        return self.__class__(self.scale ** other, d)
 
     def __truediv__(self, other):
-        if isinstance(other, string_types):
+        if isinstance(other, str):
             new = self.add(other, -1)
         elif isinstance(other, Number):
             new = self.copy()
@@ -544,7 +678,7 @@ class ParserHelper(UnitsContainer):
 
     def __rtruediv__(self, other):
         new = self.__pow__(-1)
-        if isinstance(other, string_types):
+        if isinstance(other, str):
             new = new.add(other, 1)
         elif isinstance(other, Number):
             new.scale *= other
@@ -557,25 +691,28 @@ class ParserHelper(UnitsContainer):
 
 
 #: List of regex substitution pairs.
-_subs_re = [('\N{DEGREE SIGN}', " degree"),
-            (r"([\w\.\-\+\*\\\^])\s+", r"\1 "), # merge multiple spaces
-            (r"({0}) squared", r"\1**2"),  # Handle square and cube
-            (r"({0}) cubed", r"\1**3"),
-            (r"cubic ({0})", r"\1**3"),
-            (r"square ({0})", r"\1**2"),
-            (r"sq ({0})", r"\1**2"),
-            (r"\b([0-9]+\.?[0-9]*)(?=[e|E][a-zA-Z]|[a-df-zA-DF-Z])", r"\1*"),  # Handle numberLetter for multiplication
-            (r"([\w\.\-])\s+(?=\w)", r"\1*"),  # Handle space for multiplication
-            ]
+_subs_re = [
+    ("\N{DEGREE SIGN}", " degree"),
+    (r"([\w\.\-\+\*\\\^])\s+", r"\1 "),  # merge multiple spaces
+    (r"({}) squared", r"\1**2"),  # Handle square and cube
+    (r"({}) cubed", r"\1**3"),
+    (r"cubic ({})", r"\1**3"),
+    (r"square ({})", r"\1**2"),
+    (r"sq ({})", r"\1**2"),
+    (
+        r"\b([0-9]+\.?[0-9]*)(?=[e|E][a-zA-Z]|[a-df-zA-DF-Z])",
+        r"\1*",
+    ),  # Handle numberLetter for multiplication
+    (r"([\w\.\-])\s+(?=\w)", r"\1*"),  # Handle space for multiplication
+]
 
-#: Compiles the regex and replace {0} by a regex that matches an identifier.
+#: Compiles the regex and replace {} by a regex that matches an identifier.
 _subs_re = [(re.compile(a.format(r"[_a-zA-Z][_a-zA-Z0-9]*")), b) for a, b in _subs_re]
-_pretty_table = maketrans('⁰¹²³⁴⁵⁶⁷⁸⁹·⁻', '0123456789*-')
+_pretty_table = str.maketrans("⁰¹²³⁴⁵⁶⁷⁸⁹·⁻", "0123456789*-")
 _pretty_exp_re = re.compile(r"⁻?[⁰¹²³⁴⁵⁶⁷⁸⁹]+(?:\.[⁰¹²³⁴⁵⁶⁷⁸⁹]*)?")
 
 
 def string_preprocessor(input_string):
-
     input_string = input_string.replace(",", "")
     input_string = input_string.replace(" per ", "/")
 
@@ -584,7 +721,7 @@ def string_preprocessor(input_string):
 
     # Replace pretty format characters
     for pretty_exp in _pretty_exp_re.findall(input_string):
-        exp = '**' + pretty_exp.translate(_pretty_table)
+        exp = "**" + pretty_exp.translate(_pretty_table)
         input_string = input_string.replace(pretty_exp, exp)
     input_string = input_string.translate(_pretty_table)
 
@@ -594,38 +731,95 @@ def string_preprocessor(input_string):
 
 
 def _is_dim(name):
-    return name[0] == '[' and name[-1] == ']'
+    return name[0] == "[" and name[-1] == "]"
 
 
-class SharedRegistryObject(object):
-    """Base class for object keeping a refrence to the registree.
+class SharedRegistryObject:
+    """Base class for object keeping a reference to the registree.
 
-    Such object are for now _Quantity and _Unit, in a number of places it is
+    Such object are for now Quantity and Unit, in a number of places it is
     that an object from this class has a '_units' attribute.
 
+    Parameters
+    ----------
+
+    Returns
+    -------
+
     """
 
+    def __new__(cls, *args, **kwargs):
+        inst = object.__new__(cls)
+        if not hasattr(cls, "_REGISTRY"):
+            # Base class, not subclasses dynamically by
+            # UnitRegistry._init_dynamic_classes
+            from . import _APP_REGISTRY
+
+            inst._REGISTRY = _APP_REGISTRY
+        return inst
+
     def _check(self, other):
         """Check if the other object use a registry and if so that it is the
         same registry.
 
-        Return True is both use a registry and they use the same, False is
-        other don't use a registry and raise ValueError if other don't use the
-        same unit registry.
+        Parameters
+        ----------
+        other :
+
+
+        Returns
+        -------
+        type
+            other don't use a registry and raise ValueError if other don't use the
+            same unit registry.
 
         """
-        if self._REGISTRY is getattr(other, '_REGISTRY', None):
+        if self._REGISTRY is getattr(other, "_REGISTRY", None):
             return True
 
         elif isinstance(other, SharedRegistryObject):
-            mess = 'Cannot operate with {0} and {1} of different registries.'
-            raise ValueError(mess.format(self.__class__.__name__,
-                                         other.__class__.__name__))
+            mess = "Cannot operate with {} and {} of different registries."
+            raise ValueError(
+                mess.format(self.__class__.__name__, other.__class__.__name__)
+            )
         else:
             return False
 
+
+class PrettyIPython:
+    """Mixin to add pretty-printers for IPython"""
+
+    def _repr_html_(self):
+        if "~" in self.default_format:
+            return "{:~H}".format(self)
+        else:
+            return "{:H}".format(self)
+
+    def _repr_latex_(self):
+        if "~" in self.default_format:
+            return "${:~L}$".format(self)
+        else:
+            return "${:L}$".format(self)
+
+    def _repr_pretty_(self, p, cycle):
+        if "~" in self.default_format:
+            p.text("{:~P}".format(self))
+        else:
+            p.text("{:P}".format(self))
+
+
 def to_units_container(unit_like, registry=None):
-    """ Convert a unit compatible type to a UnitsContainer.
+    """Convert a unit compatible type to a UnitsContainer.
+
+    Parameters
+    ----------
+    unit_like :
+
+    registry :
+         (Default value = None)
+
+    Returns
+    -------
 
     """
     mro = type(unit_like).mro()
@@ -633,7 +827,7 @@ def to_units_container(unit_like, registry=None):
         return unit_like
     elif SharedRegistryObject in mro:
         return unit_like._units
-    elif string_types in mro:
+    elif str in mro:
         if registry:
             return registry._parse_units(unit_like)
         else:
@@ -643,31 +837,50 @@ def to_units_container(unit_like, registry=None):
 
 
 def infer_base_unit(q):
-    """Return UnitsContainer of q with all prefixes stripped."""
+    """
+
+    Parameters
+    ----------
+    q :
+
+
+    Returns
+    -------
+    type
+
+
+    """
     d = udict()
-    parse = q._REGISTRY.parse_unit_name
     for unit_name, power in q._units.items():
-        completely_parsed_unit = list(parse(unit_name))[-1]
-
-        _, base_unit, __ = completely_parsed_unit
+        candidates = q._REGISTRY.parse_unit_name(unit_name)
+        assert len(candidates) == 1
+        _, base_unit, _ = candidates[0]
         d[base_unit] += power
-    return UnitsContainer(dict((k, v) for k, v in d.items() if v != 0))  # remove values that resulted in a power of 0
+
+    # remove values that resulted in a power of 0
+    return UnitsContainer({k: v for k, v in d.items() if v != 0})
 
 
-def fix_str_conversions(cls):
-    """Enable python2/3 compatible behaviour for __str__."""
-    def __bytes__(self):
-        return self.__unicode__().encode(locale.getpreferredencoding())
-    cls.__unicode__ = __unicode__ = cls.__str__
-    cls.__bytes__ = __bytes__
-    if sys.version_info[0] == 2:
-        cls.__str__ = __bytes__
-    else:
-        cls.__str__ = __unicode__
-    return cls
+def getattr_maybe_raise(self, item):
+    """Helper function to invoke at the beginning of all overridden ``__getattr__``
+    methods. Raise AttributeError if the user tries to ask for a _ or __ attribute.
+
+    Parameters
+    ----------
+    item :
 
 
-class SourceIterator(object):
+    Returns
+    -------
+
+    """
+    # Double-underscore attributes are tricky to detect because they are
+    # automatically prefixed with the class name - which may be a subclass of self
+    if item.startswith("_") or item.endswith("__"):
+        raise AttributeError("%r object has no attribute %r" % (self, item))
+
+
+class SourceIterator:
     """Iterator to facilitate reading the definition files.
 
     Accepts any sequence (like a list of lines, a file or another SourceIterator)
@@ -696,10 +909,10 @@ class SourceIterator(object):
         return self
 
     def __next__(self):
-        line = ''
-        while not line or line.startswith('#'):
+        line = ""
+        while not line or line.startswith("#"):
             lineno, line = next(self.internal)
-            line = line.split('#', 1)[0].strip()
+            line = line.split("#", 1)[0].strip()
 
         self.last = lineno, line
         return lineno, line
@@ -707,8 +920,7 @@ class SourceIterator(object):
     next = __next__
 
     def block_iter(self):
-        """Iterate block including header.
-        """
+        """Iterate block including header."""
         return BlockIterator(self)
 
 
@@ -730,11 +942,52 @@ class BlockIterator(SourceIterator):
             return self.last
 
         lineno, line = SourceIterator.__next__(self)
-        if line.startswith('@end'):
+        if line.startswith("@end"):
             raise StopIteration
-        elif line.startswith('@'):
-            raise DefinitionSyntaxError('cannot nest @ directives', lineno=lineno)
+        elif line.startswith("@"):
+            raise DefinitionSyntaxError("cannot nest @ directives", lineno=lineno)
 
         return lineno, line
 
     next = __next__
+
+
+def iterable(y):
+    """Check whether or not an object can be iterated over.
+
+    Vendored from numpy under the terms of the BSD 3-Clause License. (Copyright
+    (c) 2005-2019, NumPy Developers.)
+
+    Parameters
+    ----------
+    value :
+        Input object.
+    type :
+        object
+    y :
+
+    """
+    try:
+        iter(y)
+    except TypeError:
+        return False
+    return True
+
+
+def sized(y):
+    """Check whether or not an object has a defined length.
+
+    Parameters
+    ----------
+    value :
+        Input object.
+    type :
+        object
+    y :
+
+    """
+    try:
+        len(y)
+    except TypeError:
+        return False
+    return True
